from io import StringIO
import unittest
import pandas as pd

from pandas_schema import Schema, Column
from pandas_schema.validation import LeadingWhitespaceValidation


class UnorderedSchema(unittest.TestCase):
    schema = Schema([
        Column('a'),
        Column('b', [LeadingWhitespaceValidation()])
    ], ordered=False)

    def test_fields(self):
        self.assertEqual(len(self.schema.columns), 2, 'The schema is not storing all of its columns')
        self.assertEqual(self.schema.ordered, False, 'The schema is not storing the correct value of ordered')

    def test_validate_valid(self):
        df = pd.DataFrame({
            'a': ['1', '2', '3'],
            'b': ['1', '2', '3']
        })
        results = self.schema.validate(df)
        self.assertEqual(len(results), 0, 'A correct data frame should have no errors')

    def test_validate_invalid(self):
        df = pd.DataFrame({
            'a': [' 1', '2', '3'],
            'b': [' 1', '2', '3']
        })
        results = self.schema.validate(df)
        self.assertEqual(len(results), 1, 'An incorrect data frame should report errors')

    def test_mixed_columns(self):
        """
        Tests that when ordered=False, the schema columns are associated with data frame columns by name, not position.
        In this case, the schema's column order is [a, b], while the data frame's order is [b, a]. There is an error in
        column b in the data frame (leading whitespace), and a validation on column b in the schema.

        Schema         a                b (validation)
        Data Frame     b (error)        a

        Thus there will only be an error if column b in the schema is linked to column b in the data frame,
        as is correct behaviour.
        """

        df = pd.read_csv(StringIO('''
b,a
 1,1
2,3
3,3
        '''), sep=',', header=0, dtype=str)
        results = self.schema.validate(df)

        self.assertEqual(len(results), 1, 'There should be 1 error')
        self.assertEqual(results[0].row, 0)
        self.assertEqual(results[0].column, 'b',
                         'The Schema object is not associating columns and column schemas by name')


class OrderedSchema(unittest.TestCase):
    schema = Schema([
        Column('a', [LeadingWhitespaceValidation()]),
        Column('b')
    ], ordered=True)

    def test_mixed_columns(self):
        """
        Tests that when ordered=True, the schema columns are associated with data frame columns by position, not name.

        In this case, the schema's column order is [a, b], while the data frame's order is [b, a]. There is an error in
        column b in the data frame (leading whitespace), and a validation on column a in the schema.

        Schema         a (validation)   b
        Data Frame     b (error)        a

        Thus there will only be an error if column b in the schema is linked to column a in the data frame,
        as is correct behaviour when ordered=True.
        """
        df = pd.read_csv(StringIO('''
b,a
 1,1
2,3
3,3
        '''), sep=',', header=0, dtype=str)
        results = self.schema.validate(df)

        self.assertEqual(len(results), 1, 'There should be 1 error')
        self.assertEqual(results[0].row, 0)
        self.assertEqual(results[0].column, 'b',
                         'The Schema object is not associating columns and column schemas by position')



import pandas as pd
import typing

from .errors import PanSchInvalidSchemaError
from .validation_warning import ValidationWarning
from .column import Column


class Schema:
    """
    A schema that defines the columns required in the target DataFrame
    """

    def __init__(self, columns: typing.Iterable[Column], ordered: bool = False):
        """
        :param columns: A list of column objects
        :param ordered: True if the Schema should associate its Columns with DataFrame columns by position only, ignoring
            the header names. False if the columns should be associated by column header names only. Defaults to False
        """
        if not columns:
            raise PanSchInvalidSchemaError('An instance of the schema class must have a columns list')

        if not isinstance(columns, typing.List[Column]):
            raise PanSchInvalidSchemaError('The columns field must be a list of Column objects')

        if not isinstance(ordered, bool):
            raise PanSchInvalidSchemaError('The ordered field must be a boolean')

        self.columns = list(columns)
        self.ordered = ordered

    def validate(self, df: pd.DataFrame) -> typing.List[ValidationWarning]:
        """
        Runs a full validation of the target DataFrame using the internal columns list

        :param df: A pandas DataFrame to validate
        :return: A list of ValidationWarning objects that list the ways in which the DataFrame was invalid
        """
        errors = []

        # It's an error if the number of columns in the schema and data frame are different
        df_cols = len(df.columns)
        schema_cols = len(self.columns)
        if df_cols != schema_cols:
            errors.append(ValidationWarning(
                'Invalid number of columns. The schema specifies {}, but the data frame has {}'.format(schema_cols,
                                                                                                       df_cols)))
            return errors

        # We associate the column objects in the schema with data frame series either by name or by position, depending
        # on the value of self.ordered
        if self.ordered:
            series = [x[1] for x in df.iteritems()]
            column_pairs = zip(series, self.columns)
        else:
            column_pairs = []
            for column in self.columns:

                # Throw an error if the schema column isn't in the data frame
                if column.name not in df:
                    errors.append(ValidationWarning(
                        'The column {} exists in the schema but not in the data frame'.format(column.name)))
                    return errors

                column_pairs.append((df[column.name], column))

        # Iterate over each pair of schema columns and data frame series and run validations
        for series, column in column_pairs:
            errors += column.validate(series)

        return sorted(errors, key=lambda e: e.row)



class ValidationWarning:
    """
    Represents a difference between the schema and data frame, found during the validation of the data frame
    """

    def __init__(self, message: str, value: str = None, row: int = None, column: str = None):
        self.message = message
        self.value = value
        """The value of the failing cell in the DataFrame"""
        self.row = row
        """The row index (usually an integer starting from 0) of the cell that failed the validation"""
        self.column = column
        """The column name of the cell that failed the validation"""

    def __str__(self) -> str:
        """
        The entire warning message as a string
        """
        if self.row is not None and self.column is not None and self.value is not None:
            return '{{row: {}, column: "{}"}}: "{}" {}'.format(self.row, self.column, self.value, self.message)
        else:
            return self.message



import unittest

from pandas_schema import ValidationWarning


class WarningTest(unittest.TestCase):
    MESSAGE = 'Message'

    def test_shows_row_col(self):
        """
        Checks that a validation warning shows the row and column even if they are falsy values
        :return:
        """
        v = ValidationWarning(self.MESSAGE, '', 0, '')
        vs = str(v)
        self.assertRegexpMatches(vs, 'row')
        self.assertRegexpMatches(vs, self.MESSAGE)
        self.assertRegexpMatches(vs, 'column')

    def test_no_row_col(self):
        """
        Checks that a validation warning doesn't mention the row or column if neither is specified
        :return:
        """
        v = ValidationWarning(self.MESSAGE)
        vs = str(v)
        self.assertNotRegexpMatches(vs, 'row')
        self.assertRegexpMatches(vs, self.MESSAGE)
        self.assertNotRegexpMatches(vs, 'column')



import unittest
import pandas as pd

from pandas_schema import Column
from pandas_schema.validation import CanConvertValidation, LeadingWhitespaceValidation, TrailingWhitespaceValidation


class SingleValidationColumn(unittest.TestCase):
    """
    Test a column with one single validation
    """
    NAME = 'col1'

    col = Column(NAME, [CanConvertValidation(int)], allow_empty=False)
    ser = pd.Series([
        'a',
        'b',
        'c'
    ])

    def test_name(self):
        self.assertEqual(self.col.name, self.NAME, 'A Column does not store its name correctly')

    def test_outputs(self):
        results = self.col.validate(self.ser)

        self.assertEqual(len(results), len(self.ser), 'A Column produces the wrong number of errors')
        for i in range(2):
            self.assertTrue(any([r.row == i for r in results]), 'A Column does not report errors for every row')


class DoubleValidationColumn(unittest.TestCase):
    """
    Test a column with two different validations
    """
    NAME = 'col1'

    col = Column(NAME, [TrailingWhitespaceValidation(), LeadingWhitespaceValidation()], allow_empty=False)
    ser = pd.Series([
        ' a ',
        ' b ',
        ' c '
    ])

    def test_outputs(self):
        results = self.col.validate(self.ser)

        # There should be 6 errors, 2 for each row
        self.assertEqual(len(results), 2 * len(self.ser), 'A Column produces the wrong number of errors')
        for i in range(2):
            in_row = [r for r in results if r.row == i]
            self.assertEquals(len(in_row), 2, 'A Column does not report both errors for every row')


class AllowEmptyColumn(unittest.TestCase):
    """
    Test a column with one single validation that allows empty columns
    """
    NAME = 'col1'

    col = Column(NAME, [CanConvertValidation(int)], allow_empty=True)
    ser = pd.Series([
        '',
    ])

    def test_outputs(self):
        results = self.col.validate(self.ser)
        self.assertEqual(len(results), 0, 'allow_empty is not allowing empty columns')



import typing
import pandas as pd

from . import validation
from .validation_warning import ValidationWarning

class Column:
    def __init__(self, name: str, validations: typing.Iterable['validation._BaseValidation'] = [], allow_empty=False):
        """
        Creates a new Column object

        :param name: The column header that defines this column. This must be identical to the header used in the CSV/Data Frame you are validating.
        :param validations: An iterable of objects implementing _BaseValidation that will generate ValidationErrors
        :param allow_empty: True if an empty column is considered valid. False if we leave that logic up to the Validation
        """
        self.name = name
        self.validations = list(validations)
        self.allow_empty = allow_empty

    def validate(self, series: pd.Series) -> typing.List[ValidationWarning]:
        """
        Creates a list of validation errors using the Validation objects contained in the Column

        :param series: A pandas Series to validate
        :return: An iterable of ValidationError instances generated by the validation
        """
        return [error for validation in self.validations for error in validation.get_errors(series, self)]
